"""Supermercado.super_utils but for other TMS.

This submodule is adapted from mapbox/supermercado project:

The MIT License (MIT)

Copyright (c) 2015 Mapbox

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""
import json
import re
from typing import Any, Dict, Generator, List, Sequence, Tuple

import attr
import morecantile
import numpy
import numpy.typing as npt


def parseString(tilestring, matcher) -> List[int]:
    """Parse Tile."""
    tile = [int(r) for r in matcher.match(tilestring).group().split("-")]
    tile.append(tile.pop(0))
    return tile


def get_range(xyz: npt.NDArray) -> Tuple[int, int, int, int]:
    """Get tiles extrema."""
    return xyz[:, 0].min(), xyz[:, 0].max(), xyz[:, 1].min(), xyz[:, 1].max()


def burnXYZs(
    tiles: npt.NDArray, xmin: float, xmax: float, ymin: float, ymax: float, pad: int = 1
):
    """Burn XYZ tiles."""
    # make an array of shape (xrange + 3, yrange + 3)
    burn = numpy.zeros(
        (xmax - xmin + (pad * 2 + 1), ymax - ymin + (pad * 2 + 1)), dtype=bool
    )

    # using the tile xys as indicides, burn in True where a tile exists
    burn[(tiles[:, 0] - xmin + pad, tiles[:, 1] - ymin + pad)] = True

    return burn


def tile_parser(tiles: npt.ArrayLike, parsenames: bool = False) -> numpy.ndarray:
    """Parse Tile."""
    if parsenames:
        tMatch = re.compile(r"[\d]+-[\d]+-[\d]+")
        tiles = numpy.array([parseString(t, tMatch) for t in tiles])
    else:
        tiles = numpy.array([json.loads(t) for t in tiles])

    return tiles


def get_idx() -> numpy.ndarray:
    """Get numpy identity matrix."""
    tt = numpy.zeros((3, 3), dtype=bool)
    tt[1, 1] = True
    return numpy.dstack(numpy.where(~tt))[0] - 1


def get_zoom(tiles: npt.NDArray) -> int:
    """Get Zoom."""
    t, d = tiles.shape
    if t < 1 or d != 3:
        raise ValueError("Tiles must be of shape n, 3")

    if tiles[:, 2].min() != tiles[:, 2].max():
        raise ValueError("All tile zooms must be the same")

    return tiles[0, 2]


def filter_features(  # noqa: C901
    features: Sequence[Dict[Any, Any]]
) -> Generator[Dict, None, None]:
    """Filter feature."""
    for f in features:
        if "geometry" in f and "type" in f["geometry"]:
            if f["geometry"]["type"] == "Polygon":
                yield f

            elif f["geometry"]["type"] == "Point":
                yield f

            elif f["geometry"]["type"] == "LineString":
                yield f

            elif f["geometry"]["type"] == "MultiPolygon":
                for part in f["geometry"]["coordinates"]:
                    yield {
                        "type": "Feature",
                        "geometry": {"type": "Polygon", "coordinates": part},
                    }

            elif f["geometry"]["type"] == "MultiPoint":
                for part in f["geometry"]["coordinates"]:
                    yield {
                        "type": "Feature",
                        "geometry": {"type": "Point", "coordinates": part},
                    }

            elif f["geometry"]["type"] == "MultiLineString":
                for part in f["geometry"]["coordinates"]:
                    yield {
                        "type": "Feature",
                        "geometry": {"type": "LineString", "coordinates": part},
                    }


@attr.s
class Unprojecter:
    """Convert feature from TMS crs to the TMS's geographic CRS."""

    tms: morecantile.TileMatrixSet = attr.ib()

    def xy_to_lng_lat(
        self, coordinates: Sequence[Tuple[float, float]]
    ) -> numpy.ndarray:
        """Convert coordinates."""
        for c in coordinates:
            tc = numpy.array(c)
            yield numpy.dstack(
                [*self.tms._to_geographic.transform(tc[:, 0], tc[:, 1])]
            )[0].tolist()

    def unproject(self, feature: Dict[Any, Any]) -> Dict[Any, Any]:
        """Apply reprojection."""
        feature["coordinates"] = list(self.xy_to_lng_lat(feature["coordinates"]))
        return feature



"""Tiles heatmap but for other TMS.

adapted from https://github.com/mapbox/supermercado/pull/53
"""

from typing import Dict, List

import attr
import morecantile
import numpy
import numpy.typing as npt
from affine import Affine
from rasterio import features

from supermorecado import super_utils as sutils

WEB_MERCATOR_TMS = morecantile.tms.get("WebMercatorQuad")


def create_density(
    tiles: npt.NDArray, xmin: int, xmax: int, ymin: int, ymax: int, pad: int = 1
) -> numpy.ndarray:
    """Given an ndarray or tiles and range, create a density raster of tile frequency."""
    burn = numpy.zeros(
        (xmax - xmin + (pad * 2 + 1), ymax - ymin + (pad * 2 + 1)), dtype=numpy.uint32
    )
    tiles, counts = numpy.unique(tiles, return_counts=True, axis=0)
    burn[(tiles[:, 0] - xmin + pad, tiles[:, 1] - ymin + pad)] = counts
    return burn


@attr.s
class densityTiles:
    """heatmap."""

    tms: morecantile.TileMatrixSet = attr.ib(default=WEB_MERCATOR_TMS)

    def heatmap(self, tiles: npt.NDArray) -> List[Dict]:
        """Creates a vector "heatmap" of tile densities"""
        xmin, xmax, ymin, ymax = sutils.get_range(tiles)

        zoom = sutils.get_zoom(tiles)

        burn = create_density(tiles, xmin, xmax, ymin, ymax, 0)
        density = numpy.asarray(
            numpy.flipud(numpy.rot90(burn)).astype(numpy.uint8),
            order="C",
        )

        nw = self.tms.xy(*self.tms.ul(xmin, ymin, zoom))
        se = self.tms.xy(*self.tms.ul(xmax + 1, ymax + 1, zoom))
        afftrans = Affine(
            ((se[0] - nw[0]) / float(xmax - xmin + 1)),
            0.0,
            nw[0],
            0.0,
            -((nw[1] - se[1]) / float(ymax - ymin + 1)),
            nw[1],
        )

        unprojecter = sutils.Unprojecter(self.tms)

        return [
            {
                "geometry": unprojecter.unproject(feature),
                "properties": {"n": count},
                "type": "Feature",
            }
            for feature, count in features.shapes(
                density,
                mask=density > 0,
                transform=afftrans,
            )
        ]



"""Supermercado.burntiles but for other TMS.

This submodule is adapted from mapbox/supermercado project:

The MIT License (MIT)

Copyright (c) 2015 Mapbox

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

from typing import Any, Dict, Sequence, Tuple

import attr
import morecantile
import numpy
from affine import Affine
from rasterio import features

WEB_MERCATOR_TMS = morecantile.tms.get("WebMercatorQuad")


def _feature_extrema(geometry: Dict) -> Tuple[float, float, float, float]:
    """Get bounds extrema for a feature."""
    if geometry["type"] == "Polygon":
        x, y = zip(*[c for part in geometry["coordinates"] for c in part])

    elif geometry["type"] == "MultiPolygon":
        x, y = zip(
            *[c for poly in geometry["coordinates"] for part in poly for c in part]
        )

    elif geometry["type"] == "LineString":
        x, y = zip(*list(geometry["coordinates"]))

    elif geometry["type"] == "Point":
        x, y = geometry["coordinates"]
        return x, y, x, y

    return min(x), min(y), max(x), max(y)


@attr.s
class burnTiles:
    """Burntiles."""

    tms: morecantile.TileMatrixSet = attr.ib(default=WEB_MERCATOR_TMS)

    def find_extrema(
        self,
        features: Sequence[Dict[Any, Any]],
        truncate: bool = True,
    ) -> Tuple[float, float, float, float]:
        """Get bounds extrema for a set of features."""
        epsilon = 1.0e-10
        min_x, min_y, max_x, max_y = zip(
            *[_feature_extrema(f["geometry"]) for f in features]
        )

        tms_bbox = self.tms.bbox
        if truncate:
            return (
                max(min(min_x) + epsilon, tms_bbox.left),
                max(min(min_y) + epsilon, tms_bbox.bottom),
                min(max(max_x) - epsilon, tms_bbox.right),
                min(max(max_y) - epsilon, tms_bbox.top),
            )

        else:
            return (
                min(min_x) + epsilon,
                min(min_y) + epsilon,
                max(max_x) - epsilon,
                max(max_y) - epsilon,
            )

    def project_geom(self, geom: Dict) -> Dict:
        """reproject geom in TMS CRS."""
        if geom["type"] == "Polygon":
            return {
                "type": geom["type"],
                "coordinates": [
                    [self.tms.xy(*coords) for coords in part]
                    for part in geom["coordinates"]
                ],
            }

        elif geom["type"] == "LineString":
            return {
                "type": geom["type"],
                "coordinates": [self.tms.xy(*coords) for coords in geom["coordinates"]],
            }

        elif geom["type"] == "Point":
            return {
                "type": geom["type"],
                "coordinates": self.tms.xy(*geom["coordinates"]),
            }

        elif geom["type"] == "MultiPolygon":
            return {
                "type": geom["type"],
                "coordinates": [
                    [[self.tms.xy(*coords) for coords in part] for part in poly]
                    for poly in geom["coordinates"]
                ],
            }

        else:
            raise Exception(f"Invalid geometry type {geom['type']}")

    def tile_extrema(
        self, bounds: Tuple[float, float, float, float], zoom: int
    ) -> Dict:
        """Tiles min/max at the given zoom for bounds."""
        minimumTile = self.tms.tile(bounds[0], bounds[3], zoom)
        maximumTile = self.tms.tile(bounds[2], bounds[1], zoom)

        return {
            "x": {"min": minimumTile.x, "max": maximumTile.x + 1},
            "y": {"min": minimumTile.y, "max": maximumTile.y + 1},
        }

    def make_transform(self, tilerange: Dict, zoom: int) -> Affine:
        """Create Affine Transform from extrema."""
        ulx, uly = self.tms.xy(
            *self.tms.ul(tilerange["x"]["min"], tilerange["y"]["min"], zoom)
        )

        lrx, lry = self.tms.xy(
            *self.tms.ul(tilerange["x"]["max"], tilerange["y"]["max"], zoom)
        )

        xcell = (lrx - ulx) / float(tilerange["x"]["max"] - tilerange["x"]["min"])
        ycell = (uly - lry) / float(tilerange["y"]["max"] - tilerange["y"]["min"])

        return Affine(xcell, 0, ulx, 0, -ycell, uly)

    def burn(
        self,
        polys: Sequence[Dict[Any, Any]],
        zoom: int,
        truncate: bool = True,
    ) -> numpy.ndarray:
        """Burn geometries to Tiles."""
        bounds = self.find_extrema(polys, truncate=truncate)

        tilerange = self.tile_extrema(bounds, zoom)
        afftrans = self.make_transform(tilerange, zoom)

        burn = features.rasterize(
            ((self.project_geom(geom["geometry"]), 255) for geom in polys),
            out_shape=(
                (
                    tilerange["y"]["max"] - tilerange["y"]["min"],
                    tilerange["x"]["max"] - tilerange["x"]["min"],
                )
            ),
            transform=afftrans,
            all_touched=True,
        )

        xys = numpy.fliplr(numpy.dstack(numpy.where(burn))[0])
        xys[:, 0] += tilerange["x"]["min"]
        xys[:, 1] += tilerange["y"]["min"]

        return numpy.append(
            xys, numpy.zeros((xys.shape[0], 1), dtype=numpy.uint8) + zoom, axis=1
        )



"""
supermorecado is an adaptation of mapbox/supermorecado to work with other TMS grid.

"""

__version__ = "0.1.0"

from .burntiles import burnTiles  # noqa
from .densitytiles import densityTiles  # noqa
from .edge_finder import findedges  # noqa
from .uniontiles import unionTiles  # noqa



"""Morecantile command line interface"""

import json

import click
import cligj
import morecantile

from supermorecado import burnTiles, densityTiles, findedges
from supermorecado import super_utils as sutils
from supermorecado import unionTiles


@click.group(help="Command line interface for the Supermorecado Python package.")
def cli():
    """Supermorecado CLI."""
    pass


@cli.command(
    short_help="For a stream of [<x>, <y>, <z>] tiles, return only those tiles that are on the edge."
)
@click.argument("inputtiles", default="-", required=False)
@click.option("--parsenames", is_flag=True)
def edges(inputtiles, parsenames):
    """
    For a stream of [<x>, <y>, <z>] tiles, return only those tiles that are on the edge.
    """
    try:
        inputtiles = click.open_file(inputtiles).readlines()
    except IOError:
        inputtiles = [inputtiles]

    # parse the input stream into an array
    tiles = sutils.tile_parser(inputtiles, parsenames)
    for t in findedges(tiles):
        click.echo(t.tolist())


@cli.command(
    short_help="Returns the unioned shape of a stream of [<x>, <y>, <z>] tiles in GeoJSON."
)
@click.argument("inputtiles", default="-", required=False)
@click.option("--parsenames", is_flag=True)
@click.option(
    "--identifier",
    type=click.Choice(morecantile.tms.list()),
    default="WebMercatorQuad",
    help="TileMatrixSet identifier.",
)
@click.option(
    "--tms",
    help="Path to TileMatrixSet JSON file.",
    type=click.Path(),
)
def union(inputtiles, parsenames, identifier, tms):
    """
    Returns the unioned shape of a stream of [<x>, <y>, <z>] tiles in GeoJSON.
    """
    try:
        inputtiles = click.open_file(inputtiles).readlines()
    except IOError:
        inputtiles = [inputtiles]

    tiles = sutils.tile_parser(inputtiles, parsenames)

    tilematrixset = morecantile.tms.get(identifier)
    if tms:
        with open(tms, "r") as f:
            tilematrixset = morecantile.TileMatrixSet(**json.load(f))

    uniontiles = unionTiles(tms=tilematrixset)
    for u in uniontiles.union(tiles):
        click.echo(json.dumps(u))


@cli.command(
    short_help="Burn a stream of GeoJSONs into a output stream of the tiles they intersect for a given zoom."
)
@cligj.features_in_arg
@cligj.sequence_opt
@click.argument("zoom", type=int)
@click.option(
    "--identifier",
    type=click.Choice(morecantile.tms.list()),
    default="WebMercatorQuad",
    help="TileMatrixSet identifier.",
)
@click.option(
    "--tms",
    help="Path to TileMatrixSet JSON file.",
    type=click.Path(),
)
def burn(features, sequence, zoom, identifier, tms):
    """
    Burn a stream of GeoJSONs into a output stream of the tiles they intersect for a given zoom.
    """
    features = list(sutils.filter_features(features))

    tilematrixset = morecantile.tms.get(identifier)
    if tms:
        with open(tms, "r") as f:
            tilematrixset = morecantile.TileMatrixSet(**json.load(f))

    burntiles = burnTiles(tms=tilematrixset)
    for t in burntiles.burn(features, zoom):
        click.echo(t.tolist())


@cli.command(short_help="Creates a vector `heatmap` of tile densities.")
@click.argument("inputtiles", default="-", required=False)
@click.option("--parsenames", is_flag=True)
@click.option(
    "--identifier",
    type=click.Choice(morecantile.tms.list()),
    default="WebMercatorQuad",
    help="TileMatrixSet identifier.",
)
@click.option(
    "--tms",
    help="Path to TileMatrixSet JSON file.",
    type=click.Path(),
)
def heatmap(inputtiles, parsenames, identifier, tms):
    """
    Creates a vector `heatmap` of tile densities.
    """
    try:
        inputtiles = click.open_file(inputtiles).readlines()
    except IOError:
        inputtiles = [inputtiles]

    tiles = sutils.tile_parser(inputtiles, parsenames)

    tilematrixset = morecantile.tms.get(identifier)
    if tms:
        with open(tms, "r") as f:
            tilematrixset = morecantile.TileMatrixSet(**json.load(f))

    tiledensity = densityTiles(tms=tilematrixset)
    for u in tiledensity.heatmap(tiles):
        click.echo(json.dumps(u))
